/**
 * @file
 * Provides GridStack loader.
 */

(function ($, Drupal, drupalSettings, _, _db, window) {

  'use strict';

  Drupal.gridstack = Drupal.gridstack || {};

  /**
   * GridStack front-end public methods.
   *
   * @namespace
   */
  Drupal.gridstack.fe = {
    $el: null,
    $instance: null,
    horizontalMargin: 0,
    config: {},
    breakpoints: null,
    windowWidth: 0,
    options: {},
    serializedData: null,

    serialized: function (box) {
      var me = this;
      var data = _.map(box, function (grid) {
        var node = $(grid).data('_gridstack_node');
        // Bail out, iF disabled by disabling Auto option, else error.
        if (typeof node === 'undefined') {
          return {};
        }
        var grids = {
          x: node.x,
          y: node.y,
          width: node.width,
          height: node.height
        };

        return grids;
      }, me);

      return data;
    },

    resizeGrid: function () {
      var me = this;

      me.updateClasses();

      if (!_.isNull(me.breakpoints)) {
        me.updateGrid();
      }

      me.updateRatio();
    },

    updateClasses: function () {
      var me = this;
      var minWidth = me.config.minWidth;

      if (me.windowWidth <= minWidth) {
        me.$el.removeClass('gridstack--enabled').addClass('gridstack--disabled');
        me.$el.removeAttr('style');
      }
      else {
        me.$el.removeClass('gridstack--disabled').addClass('gridstack--enabled');
      }

      if (!me.config.auto) {
        me.$el.addClass('gridstack--uninit');
      }
    },

    updateRatio: function () {
      var me = this;

      // @todo https://github.com/gridstack/gridstack.js/issues/304
      // Cannot rely on gsWidth x gsHeight to have aspect ratio with margins.
      me.$el.children().each(function () {
        var $item = $(this);
        var gsWidth = $item.data('gsWidth');
        var gsHeight = $item.data('gsHeight');
        var width = (gsWidth * me.$instance.cellWidth()) + me.horizontalMargin;
        var height = ((gsHeight * me.options.cellHeight) + (me.options.verticalMargin));
        var pad = Math.round(((height / width) * 100), 2);
        var $media = $('.media--ratio', this);

        if ($media.length && width > 0 && height > 0) {
          $media.css({paddingBottom: pad + '%'});
          // @todo $item.css({height: (pad * $media.width()) / 100});
        }
      });
    },

    updateGrid: function () {
      var me = this;
      var activeGrid = null;
      var keys = _.keys(me.breakpoints);
      var max = parseInt(_.last(keys));
      var tcl = null;
      var twd = null;

      var breakpoints = keys.sort(function (a, b) {
        return (me.options.mobileFirst) ? a - b : b - a;
      });

      _.each(breakpoints, function (width, i) {
        if (me.windowWidth <= width) {
          tcl = me.breakpoints[width];
          twd = width;
        }
        else if (me.windowWidth >= max) {
          tcl = me.breakpoints[max];
          twd = max;
        }
      });

      if (!_.isNull(tcl)) {
        me.$instance.setGridWidth(tcl, true);

        // {"480":1,"767":2,"1024":3,"1400":12}.
        activeGrid = me.activeGrid(twd, max);
        if (!_.isNull(activeGrid)) {
          _.each(activeGrid, function (item, i) {
            var $box = $('> .gridstack__box:visible', me.$el).eq(i);

            item = _.isObject(item) ? _.values(item) : item;

            // Params: el, x, y, width, height.
            me.$instance.update($box, item[0], item[1], item[2], item[3]);
          });
        }
      }
    },

    activeGrid: function (width, max) {
      var me = this;
      var mw = me.$el.data('mdWidth');
      var sw = me.$el.data('smWidth');

      // Do not do anything if no responsive grids defined.
      if (sw || mw) {
        if (max <= width) {
          return me.serializedData;
        }
        else if (mw <= width) {
          return me.$el.data('mdGrids');
        }
        else if (sw <= width) {
          return me.$el.data('smGrids');
        }
      }

      return null;
    },

    resize: function () {
      var me = this;

      var doResize = function () {
        me.windowWidth = window.innerWidth || document.documentElement.clientWidth || document.body.clientWidth;
        me.resizeGrid();
      };

      // Two tasks: load the active grid per breakpoint, and update on resizing.
      _db.resize(doResize)();
    },

    cleanUp: function () {
      var me = this;

      me.$el.removeClass('grid-stack-12 gridstack--destroyed');
      me.$el.removeAttr('data-config data-sm-grids data-md-grids');
    },

    destroy: function () {
      var me = this;

      if (me.$instance !== null) {
        me.$instance.destroy(false);

        me.$el.addClass('gridstack--destroyed');
        me.$el.removeAttr('style');
      }
    },

    init: function () {
      var me = this;
      var defaults = drupalSettings.gridstack || {};
      var box = $('> .gridstack__box:visible', me.$el);
      var base = {
        mobileFirst: false,
        itemClass: 'gridstack__box',
        handle: '.box__content'
      };

      me.config = me.$el.data('config') || {};
      me.options = $.extend({}, defaults, base, me.config);

      me.$el.gridstack(me.options);

      me.$instance = me.$el.data('gridstack');
      me.breakpoints = me.$el.data('breakpoints') || null;
      me.serializedData = me.serialized(box);
      me.horizontalMargin = me.options.horizontalMargin ? 16 : 0;
      me.resize();
      me.cleanUp();
    }

  };

  /**
   * GridStack utility functions.
   *
   * @param {int} i
   *   The index of the current element.
   * @param {HTMLElement} elm
   *   The GridStack HTML element.
   */
  function doGridStack(i, elm) {
    var me = Drupal.gridstack.fe;

    me.$el = $(elm);

    me.init();
  }

  /**
   * Attaches gridstack behavior to HTML element identified by .gridstack--js.
   *
   * @type {Drupal~behavior}
   */
  Drupal.behaviors.gridstack = {
    attach: function (context) {
      $('.gridstack--js:not(.ungridstack)', context).once('gridstack').each(doGridStack);
    }
  };

})(jQuery, Drupal, drupalSettings, _, dBlazy, this);
